import lamindb as ln


def test_delete():
    labels = [ln.ULabel(name=name) for name in ["label1", "label2", "label3"]]
    ln.save(labels)
    ln.delete(labels[0])
    ln.delete([labels[1], labels[2]])
    assert ln.ULabel.filter(name__in=["label1", "label2", "label3"]).count() == 0



import pytest
from django.db.models.deletion import ProtectedError

import lamindb as ln


def test_is_new_version_of_versioned_transform():
    # attempt to create a transform with an invalid version
    with pytest.raises(ValueError) as error:
        transform = ln.Transform(name="My transform", version=0)
    assert (
        error.exconly()
        == "ValueError: `version` parameter must be `None` or `str`, e.g., '0.1', '1',"
        " '2', etc."
    )
    # with pytest.raises(ValueError) as error:
    #     transform = ln.Transform(name="My transform", version="0")
    # assert (
    #     error.exconly()
    #     == "ValueError: Please choose a version != '0', as it could be interpreted as `None`"  # noqa
    # )

    # create a versioned transform
    transform = ln.Transform(name="My transform", version="1")
    assert transform.version == "1"

    transform.save()

    # create new transform from old transform
    transform_v2 = ln.Transform(name="My 2nd transform", is_new_version_of=transform)
    assert transform_v2.id[:12] == transform.id[:12]  # stem_id
    assert transform.version == "1"
    assert (
        transform.initial_version_id is None
    )  # initial transform has initial_version_id None
    assert transform_v2.initial_version_id == transform.id
    assert transform_v2.version == "2"

    transform_v2.save()

    # create new transform from newly versioned transform
    transform_v3 = ln.Transform(name="My transform", is_new_version_of=transform_v2)
    assert transform_v3.id[:12] == transform.id[:12]  # stem_id
    assert transform_v3.initial_version_id == transform.id
    assert transform_v3.version == "3"

    # default name
    transform_v3 = ln.Transform(is_new_version_of=transform_v2)
    assert transform_v3.name == transform_v2.name

    # wrong transform type
    with pytest.raises(TypeError) as error:
        ln.Transform(is_new_version_of=ln.ULabel(name="x"))
    assert (
        error.exconly() == "TypeError: is_new_version_of has to be of type ln.Transform"
    )

    # wrong kwargs
    with pytest.raises(ValueError) as error:
        ln.Transform(x=1)
    assert (
        error.exconly()
        == "ValueError: Only name, short_name, version, type, is_new_version_of can be"
        " passed, but you passed: {'x': 1}"
    )  # noqa

    # test that reference transform cannot be deleted
    with pytest.raises(ProtectedError):
        transform.delete()
    transform_v2.delete()
    transform_v3.delete()
    transform.delete()


def test_is_new_version_of_unversioned_transform():
    # unversioned transform
    transform = ln.Transform(name="My transform")
    assert transform.initial_version_id is None
    assert transform.version is None

    # what happens if we don't save the old transform?
    # add a test for it!
    transform.save()

    # create new transform from old transform
    new_transform = ln.Transform(name="My new transform", is_new_version_of=transform)
    assert new_transform.id[:12] == transform.id[:12]  # stem_id
    assert transform.version == "1"
    assert transform.initial_version is None
    assert new_transform.initial_version_id == transform.id
    assert new_transform.version == "2"

    transform.delete()



from inspect import signature

import anndata as ad
import lnschema_bionty as lb
import numpy as np
import pandas as pd
import pytest
from django.db.models.deletion import ProtectedError

import lamindb as ln
from lamindb import _dataset

df = pd.DataFrame({"feat1": [1, 2], "feat2": [3, 4]})

adata = ad.AnnData(
    X=np.array([[1, 2, 3], [4, 5, 6]]),
    obs=dict(feat1=["A", "B"]),
    var=pd.DataFrame(index=["MYC", "TCF7", "GATA1"]),
    obsm=dict(X_pca=np.array([[1, 2], [3, 4]])),
)

adata2 = ad.AnnData(
    X=np.array([[1, 2, 5], [4, 5, 8]]),
    obs=dict(feat1=["A", "B"]),
    var=pd.DataFrame(index=["MYC", "TCF7", "GATA1"]),
    obsm=dict(X_pca=np.array([[1, 2], [3, 4]])),
)


def test_signatures():
    # this seems currently the easiest and most transparent
    # way to test violations of the signature equality
    # the MockORM class is needed to get inspect.signature
    # to work
    class Mock:
        pass

    # class methods
    class_methods = ["from_df", "from_anndata"]
    for name in class_methods:
        setattr(Mock, name, getattr(_dataset, name))
        assert signature(getattr(Mock, name)) == _dataset.SIGS.pop(name)
    # methods
    for name, sig in _dataset.SIGS.items():
        assert signature(getattr(_dataset, name)) == sig


def test_create_delete_from_single_dataframe():
    df = ln.dev.datasets.df_iris_in_meter_study1()

    dataset = ln.Dataset.from_df(df, name="Iris flower dataset1")
    # because features weren't registered, there is no linked feature set
    assert dataset._feature_sets == {}

    # register features
    ln.save(ln.Feature.from_df(df))

    # won't work with features like so
    dataset = ln.Dataset(df, name="Iris flower dataset1")
    assert dataset._feature_sets == {}

    # will work like so
    dataset = ln.Dataset.from_df(df, name="Iris flower dataset1")
    assert "columns" in dataset._feature_sets

    dataset.save()

    # basics
    assert dataset.load().iloc[0].tolist() == df.iloc[0].tolist()
    file = dataset.file
    assert file.description == f"See dataset {dataset.id}"
    assert dataset.hash == file.hash
    assert dataset.id == file.id
    assert ln.File.filter(id=dataset.id).one_or_none() is not None
    assert ln.File.filter(id=file.id).one_or_none() is not None

    # features
    feature_list = [
        "sepal_length",
        "sepal_width",
        "petal_length",
        "petal_width",
        "iris_species_name",
    ]
    assert len(ln.Feature.filter(name__in=feature_list).list()) == 5
    feature_set = ln.FeatureSet.filter(datasets=dataset).one()
    feature_list_queried = ln.Feature.filter(feature_sets=feature_set).list()
    feature_list_queried = [feature.name for feature in feature_list_queried]
    assert set(feature_list_queried) == set(feature_list)
    # the feature_set is also linked to the file
    assert ln.FeatureSet.filter(files=dataset.file).one() == feature_set

    # accidental recreation (re-load based on hash)
    dataset1 = ln.Dataset.from_df(df, name="Iris Flower data1")
    assert dataset1.id == dataset.id
    assert dataset1.hash == dataset.hash

    # now proceed to deletion
    dataset.delete(storage=True)
    assert ln.File.filter(id=dataset.id).one_or_none() is None
    assert ln.File.filter(id=file.id).one_or_none() is None


def test_create_delete_from_single_anndata():
    ln.track(ln.Transform(name="Test transform"))
    dataset = ln.Dataset(adata, name="My adata")
    dataset.save()
    dataset.delete(storage=True)
    assert ln.File.filter(id=dataset.id).one_or_none() is None
    assert ln.File.filter(id=dataset.file.id).one_or_none() is None
    # and now with from_anndata
    lb.settings.species = "human"
    dataset = ln.Dataset.from_anndata(adata, name="My adata", field=lb.Gene.symbol)
    # let's now try passing an AnnData-like file with some feature sets linked
    ln.save(ln.Feature.from_df(adata.obs))
    file = ln.File.from_anndata(adata, description="my adata", field=lb.Gene.symbol)
    file.save()
    ln.save(lb.Gene.from_values(adata.var.index, "symbol"))
    dataset = ln.Dataset.from_anndata(file, name="My dataset", field=lb.Gene.symbol)
    dataset.save()
    dataset.describe()
    dataset.view_flow()
    feature_sets_queried = dataset.feature_sets.all()
    features_queried = ln.Feature.filter(feature_sets__in=feature_sets_queried).all()
    assert set(features_queried.list("name")) == set(adata.obs.columns)
    genes_queried = lb.Gene.filter(feature_sets__in=feature_sets_queried).all()
    assert set(genes_queried.list("symbol")) == set(adata.var.index)
    feature_sets_queried.delete()
    features_queried.delete()
    genes_queried.delete()
    dataset.delete(storage=True)
    ln.dev.run_context.run = None
    ln.dev.run_context.transform = None


def test_from_single_file():
    lb.settings.species = "human"
    ln.save(ln.Feature.from_df(adata.obs))
    file = ln.File.from_anndata(adata, description="My adata", field=lb.Gene.symbol)
    with pytest.raises(ValueError) as error:
        ln.Dataset(file)
    assert str(error.exconly()).startswith(
        "ValueError: Save file before creating dataset!"
    )
    file.save()
    with pytest.raises(ValueError) as error:
        ln.Dataset(file, file)
    assert str(error.exconly()).startswith(
        "ValueError: Only one non-keyword arg allowed: data"
    )
    transform = ln.Transform(name="My test transform")
    transform.save()
    run = ln.Run(transform)
    dataset = ln.Dataset(file, name="My new dataset", run=run)
    dataset.save()
    # test data flow
    assert dataset.run.input_files.get() == file
    # test features
    assert set(file.feature_sets.list("id")) == set(
        dataset.file.feature_sets.list("id")
    )
    assert set(file.features._feature_set_by_slot.keys()) == set(
        dataset.features._feature_set_by_slot.keys()
    )
    feature_sets_queried = file.feature_sets.all()
    features_queried = ln.Feature.filter(feature_sets__in=feature_sets_queried).all()
    feature_sets_queried.delete()
    features_queried.delete()
    dataset.delete(storage=True)
    assert ln.File.filter(id=dataset.id).one_or_none() is None
    assert ln.File.filter(id=dataset.file.id).one_or_none() is None


def test_edge_cases():
    with pytest.raises(ValueError) as error:
        ln.Dataset(df, invalid_param=1)
    assert str(error.exconly()).startswith(
        "ValueError: Only data, name, run, description, reference, reference_type can be passed, you passed: "  # noqa
    )
    with pytest.raises(ValueError) as error:
        ln.Dataset(1, name="Invalid")
    assert str(error.exconly()).startswith(
        "ValueError: Only DataFrame, AnnData and iterable of File is allowed"
    )
    file = ln.File(df, description="Test file")
    assert file._state.adding
    with pytest.raises(ValueError) as error:
        ln.Dataset([file])
    assert str(error.exconly()).startswith(
        "ValueError: Not all files are yet saved, please save them"
    )
    file.save()
    with pytest.raises(ValueError) as error:
        ln.Dataset([file, file])
    assert str(error.exconly()).startswith(
        "ValueError: Please pass files with distinct hashes: these ones are non-unique"
    )
    file.delete(storage=True)


def test_backed():
    dataset = ln.Dataset(adata, name="My test")
    dataset.backed()


def test_from_inconsistent_files():
    file1 = ln.File(df, description="My test")
    file1.save()
    file2 = ln.File(adata, description="My test2")
    file2.save()
    dataset = ln.Dataset([file1, file2], name="Inconsistent")
    dataset.save()
    with pytest.raises(RuntimeError) as error:
        dataset.load()
    assert str(error.exconly()).startswith(
        "RuntimeError: Can only load datasets where all files have the same suffix"
    )
    file1.delete(storage=True)
    file2.delete(storage=True)
    dataset.delete()


def test_from_consistent_files():
    file1 = ln.File(adata, description="My test")
    file1.save()
    file2 = ln.File(adata2, description="My test2")
    file2.save()
    transform = ln.Transform(name="My test transform")
    transform.save()
    run = ln.Run(transform)
    dataset = ln.Dataset([file1, file2], name="My test", run=run)
    dataset.save()
    assert set(dataset.run.input_files.all()) == {file1, file2}
    dataset.load()
    with pytest.raises(RuntimeError) as error:
        dataset.backed()
    assert str(error.exconly()).startswith(
        "RuntimeError: Can only call backed() for datasets with a single file"
    )
    file1.delete(storage=True)
    file2.delete(storage=True)
    dataset.delete()


def test_is_new_version_of_versioned_dataset():
    # create a versioned dataset
    dataset = ln.Dataset(df, name="test", version="1")
    assert dataset.version == "1"
    dataset.save()

    with pytest.raises(ValueError) as error:
        dataset_v2 = ln.Dataset(adata, is_new_version_of=dataset, version="1")
    assert error.exconly() == "ValueError: Please increment the previous version: '1'"

    # create new dataset from old dataset
    dataset_v2 = ln.Dataset(adata, is_new_version_of=dataset)
    assert dataset_v2.id[:18] == dataset.id[:18]  # stem_id
    assert dataset.version == "1"
    assert (
        dataset.initial_version_id is None
    )  # initial dataset has initial_version_id None
    assert dataset_v2.initial_version_id == dataset.id
    assert dataset_v2.version == "2"
    assert dataset_v2.name == "test"

    dataset_v2.save()

    # create new dataset from newly versioned dataset
    df.iloc[0, 0] = 0
    dataset_v3 = ln.Dataset(df, name="test1", is_new_version_of=dataset_v2)
    assert dataset_v3.id[:18] == dataset.id[:18]  # stem_id
    assert dataset_v3.initial_version_id == dataset.id
    assert dataset_v3.version == "3"
    assert dataset_v3.name == "test1"

    # test that reference dataset cannot be deleted
    with pytest.raises(ProtectedError):
        dataset.delete(storage=True)
    dataset_v2.delete(storage=True)
    dataset_v3.delete(storage=True)
    dataset.delete(storage=True)


def test_is_new_version_of_unversioned_dataset():
    # unversioned dataset
    dataset = ln.Dataset(df, name="test2")
    assert dataset.initial_version_id is None
    assert dataset.version is None

    # what happens if we don't save the old dataset?
    # add a test for it!
    dataset.save()

    with pytest.raises(TypeError):
        ln.Dataset(adata, is_new_version_of="wrong-type")

    # create new dataset from old dataset
    new_dataset = ln.Dataset(adata, is_new_version_of=dataset)
    assert new_dataset.id[:18] == dataset.id[:18]  # stem_id
    assert dataset.version == "1"
    assert dataset.initial_version is None
    assert new_dataset.initial_version_id == dataset.id
    assert new_dataset.version == "2"
    assert new_dataset.name == dataset.name

    dataset.delete(storage=True)



from postgrest.exceptions import APIError
from supabase.client import Client
from typing import Optional, Dict


def select_instance_by_owner_name(
    owner: str,
    name: str,
    client: Client,
) -> Optional[Dict]:
    data = (
        client.table("instance")
        .select("*, account!inner!fk_instance_account_id_account(*)")
        .eq("account.handle", owner)
        .eq("name", name)
        .execute()
        .data
    )
    if len(data) == 0:
        return None
    return data[0]


# --------------- ACCOUNT ----------------------
def sb_select_account_by_handle(
    handle: str,
    client: Client,
):
    data = client.table("account").select("*").eq("handle", handle).execute().data
    if len(data) == 0:
        return None
    return data[0]


def sb_select_account_handle_name_by_lnid(lnid: str, client: Client):
    data = (
        client.table("account").select("handle, name").eq("lnid", lnid).execute().data
    )
    if not data:
        return None
    return data[0]


# --------------- INSTANCE ----------------------
def sb_insert_instance(instance_fields: dict, client: Client):
    (
        client.table("instance")
        .insert(instance_fields, returning="minimal")
        .execute()
        .data
    )


def sb_select_instance_by_name(
    account_id: str,
    name: str,
    client: Client,
):
    data = (
        client.table("instance")
        .select("*")
        .eq("account_id", account_id)
        .eq("name", name)
        .execute()
        .data
    )
    if len(data) == 0:
        return None
    return data[0]


def sb_update_instance(instance_id: str, instance_fields: dict, client: Client):
    data = (
        client.table("instance")
        .update(instance_fields)
        .eq("id", instance_id)
        .execute()
        .data
    )
    if len(data) == 0:
        return None
    return data[0]


def sb_delete_instance(
    id: str,
    client: Client,
):
    data = client.table("instance").delete().eq("id", id).execute().data
    if len(data) == 0:
        return None
    return data[0]


# --------------- COLLABORATOR ----------------------
def sb_insert_collaborator(account_instance_fields: dict, client: Client):
    try:
        (
            client.table("account_instance")
            .insert(account_instance_fields, returning="minimal")
            .execute()
            .data
        )
    except APIError as api_error:
        pk_violation_msg = (
            'duplicate key value violates unique constraint "pk_account_instance"'
        )
        if api_error.message == pk_violation_msg:
            return "collaborator-exists-already"
        raise api_error


def sb_select_collaborator(
    instance_id: str,
    account_id: str,
    client: Client,
):
    data = (
        client.table("account_instance")
        .select("*")
        .eq("instance_id", instance_id)
        .eq("account_id", account_id)
        .execute()
        .data
    )
    if len(data) == 0:
        return None
    return data[0]


# --------------- STORAGE ----------------------
def sb_insert_storage(storage_fields: dict, client: Client):
    data = client.table("storage").insert(storage_fields).execute().data
    if len(data) == 0:
        return None
    return data[0]


def sb_select_storage(id: str, client: Client):
    data = client.table("storage").select("*").eq("id", id).execute().data
    if len(data) == 0:
        return None
    return data[0]


def sb_select_storage_by_root(root: str, client: Client):
    data = client.table("storage").select("*").eq("root", root).execute().data
    if len(data) == 0:
        return None
    return data[0]


# --------------- DBUser ----------------------
def sb_insert_db_user(instance_fields: dict, client: Client):
    try:
        data = client.table("db_user").insert(instance_fields).execute().data
    except Exception as e:
        if str(e) == str("Expecting value: line 1 column 1 (char 0)"):
            pass
        else:
            raise e
    return data[0]


def sb_select_db_user_by_instance(instance_id: str, client: Client):
    """Get the DBAccount directly associated with Instance.

    By contrast this is not the DBAccount that is linked through the
    UserInstance table.
    """
    data = (
        client.table("db_user")
        .select("*")
        .eq("instance_id", instance_id)
        .execute()
        .data
    )
    if len(data) == 0:
        return None
    return data[0]



import secrets
import string
from pathlib import Path
from typing import Optional, Union
from uuid import UUID

from pydantic import BaseModel, validator
from pydantic.networks import MultiHostDsn

from supabase import Client

from .upath import UPath


def base62(n_char: int) -> str:
    """Like nanoid without hyphen and underscore."""
    alphabet = string.digits + string.ascii_letters.swapcase()
    id = "".join(secrets.choice(alphabet) for i in range(n_char))
    return id


def secret() -> str:
    """Password or secret: 40 base62."""
    return base62(n_char=40)


def validate_schema_arg(schema: Optional[str] = None) -> str:
    if schema is None or schema == "":
        return ""
    # currently no actual validation, can add back if we see a need
    # the following just strips white spaces
    to_be_validated = [s.strip() for s in schema.split(",")]
    return ",".join(to_be_validated)


def validate_db_arg(db: Optional[str]) -> None:
    if db:
        LaminDsnModel(db=db)


def validate_unique_sqlite(
    *,
    storage_id: UUID,
    name: str,
    client: Client,
) -> None:
    # if a remote sqlite instance, make sure there is no other instance
    # that has the same name and storage location
    instances = (
        client.table("instance")
        .select("*")
        .eq("storage_id", storage_id.hex)
        .eq("name", name)
        .execute()
        .data
    )
    if len(instances) > 0:
        # retrieve account owning the first instance
        accounts = (
            client.table("account")
            .select("*")
            .eq("id", instances[0]["account_id"])
            .execute()
            .data
        )
        raise RuntimeError(
            "\nThere is already an sqlite instance with the same name and storage"
            f" location from account {accounts[0]['handle']}\nTwo sqlite instances with"
            " the same name and the same storage cannot exist\nFix: Choose another"
            f" name or load instance {accounts[0]['handle']}/{name}\nFor reference: the"
            f" instances with the same storage and name are {instances}"
        )


def get_storage_region(storage_root: Union[str, Path, UPath]) -> Optional[str]:
    storage_root_str = str(storage_root)
    if storage_root_str.startswith("s3://"):
        import botocore.session as session
        from botocore.config import Config
        from botocore.exceptions import NoCredentialsError

        bucket = storage_root_str.replace("s3://", "")
        s3_session = session.get_session()
        s3_client = s3_session.create_client("s3")
        try:
            response = s3_client.head_bucket(Bucket=bucket)
        except NoCredentialsError:  # deal with anonymous access
            s3_client = s3_session.create_client(
                "s3", config=Config(signature_version=session.UNSIGNED)
            )
            response = s3_client.head_bucket(Bucket=bucket)
        storage_region = response["ResponseMetadata"].get("HTTPHeaders", {})[
            "x-amz-bucket-region"
        ]
        # if we want to except botcore.exceptions.ClientError to reformat an
        # error message, this is how to do test for the "NoSuchBucket" error:
        #     exc.response["Error"]["Code"] == "NoSuchBucket"
    else:
        storage_region = None
    return storage_region


def validate_storage_root_arg(storage_root: str) -> None:
    if storage_root.startswith(("gs://", "s3://")):
        # check for existence happens in get_storage_region
        return None
    else:  # local path
        try:
            _ = Path(storage_root)
            return None
        except Exception:
            raise ValueError(
                "`storage` is neither a valid local, a Google Cloud nor an S3 path."
            )


def get_storage_type(storage_root: str):
    if str(storage_root).startswith("s3://"):
        return "s3"
    elif str(storage_root).startswith("gs://"):
        return "gs"
    else:
        return "local"


class LaminDsn(MultiHostDsn):
    """Custom DSN Type for Lamin.

    This class allows us to customize the allowed schemes for databases
    and also handles the parsing and building of DSN strings with the
    database name instead of URL path.
    """

    allowed_schemes = {
        "postgresql",
        # future enabled schemes
        # "snowflake",
        # "bigquery"
    }
    user_required = True
    __slots__ = ()

    @property
    def database(self):
        return self.path[1:]

    @classmethod
    def build(
        cls,
        *,
        scheme: str,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: str,
        port: Optional[str] = None,
        database: Optional[str] = None,
        query: Optional[str] = None,
        fragment: Optional[str] = None,
        **_kwargs: str,
    ) -> str:
        return super().build(
            scheme=scheme,
            user=user,
            password=password,
            host=host,
            port=port,
            path=f"/{database}",
            query=query,
            fragment=fragment,
        )


class LaminDsnModel(BaseModel):
    db: LaminDsn

    @validator("db")
    def check_db_name(cls, v):
        assert v.path and len(v.path) > 1, "database must be provided"
        return v



# Generated by Django 4.2.1 on 2023-09-07 18:56

import django.db.models.deletion
from django.db import migrations, models

import lnschema_core.users


class Migration(migrations.Migration):
    dependencies = [
        ("lnschema_core", "0017_dataset_initial_version_dataset_version"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="DatasetLabel",
            new_name="DatasetULabel",
        ),
        migrations.RenameModel(
            old_name="FileLabel",
            new_name="FileULabel",
        ),
        migrations.RenameModel(
            old_name="Label",
            new_name="ULabel",
        ),
        migrations.RenameField(
            model_name="dataset",
            old_name="labels",
            new_name="ulabels",
        ),
        migrations.RenameField(
            model_name="datasetulabel",
            old_name="label",
            new_name="ulabel",
        ),
        migrations.RenameField(
            model_name="file",
            old_name="labels",
            new_name="ulabels",
        ),
        migrations.RenameField(
            model_name="fileulabel",
            old_name="label",
            new_name="ulabel",
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="created_by",
            field=models.ForeignKey(
                default=lnschema_core.users.current_user_id,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="created_ulabels",
                to="lnschema_core.user",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="datasetulabel",
            unique_together={("dataset", "ulabel")},
        ),
        migrations.AlterUniqueTogether(
            name="fileulabel",
            unique_together={("file", "ulabel")},
        ),
    ]



import shutil
from pathlib import Path

from lamin_utils import logger

from ._close import close
from ._settings import settings
from .dev._settings_load import load_instance_settings
from .dev._settings_store import instance_settings_file


def delete(instance_name: str, force: bool = False):
    """Delete an instance."""
    if "/" in instance_name:
        logger.warning(
            "Deleting the instance of another user is currently not supported with the"
            " CLI. Please provide only the instance name when deleting an instance ('/'"
            " delimiter not allowed)."
        )
        raise ValueError("Invalid instance name: '/' delimiter not allowed.")

    if not force:
        valid_responses = ["y", "yes"]
        user_input = (
            input("Are you sure you want to delete this instance? (y/n): ")
            .strip()
            .lower()
        )
        if user_input not in valid_responses:
            return -1

    instance_identifier = f"{settings.user.handle}/{instance_name}"
    logger.info(f"deleting instance {instance_identifier}")
    settings_file = instance_settings_file(instance_name, settings.user.handle)
    if not settings_file.exists():
        logger.warning(
            "could not delete as instance settings do not exist locally. did you"
            " provide a wrong instance name? could you try loading it?"
        )
        return None
    isettings = load_instance_settings(settings_file)

    delete_settings(settings_file)
    if settings._instance_exists:
        if instance_identifier == settings.instance.identifier:
            close(mute=True)  # close() does further operations, unlocking...
            settings._instance_settings = None
    delete_cache(isettings.storage.cache_dir)
    if isettings.dialect == "sqlite":
        if isettings._sqlite_file.exists():
            isettings._sqlite_file.unlink()
            logger.success("    deleted '.lndb' sqlite file")
        else:
            logger.warning("    '.lndb' sqlite file does not exist")
    if isettings.is_remote:
        logger.warning(
            "    consider manually deleting your remote instance on lamin.ai"
        )
    logger.warning(
        f"    consider manually deleting your stored data: {isettings.storage.root}"
    )


def delete_cache(cache_dir: Path):
    if cache_dir is not None and cache_dir.exists():
        shutil.rmtree(cache_dir)
        logger.success("    instance cache deleted")


def delete_settings(settings_file: Path):
    if settings_file.exists():
        settings_file.unlink()
        logger.success(f"    deleted instance settings file: {settings_file}")
    else:
        logger.warning(f"    instance settings file doesn't exist: {settings_file}")



# flake8: noqa
import builtins
import os

from lamin_utils import logger

from ._settings_instance import InstanceSettings

IS_RUN_FROM_IPYTHON = getattr(builtins, "__IPYTHON__", False)
IS_SETUP = False
MISSING_MIGRATIONS_WARNING = """

Your database is not up to date with your installed Python library.

The database misses the following migrations:
{missing_migrations}

Only if you are an admin and manage migrations manually, deploy them to the database: lamin migrate deploy

Otherwise, downgrade your Python library to match the database!
"""


def get_migrations_to_sync():
    from django.db import DEFAULT_DB_ALIAS, connections
    from django.db.migrations.executor import MigrationExecutor

    connection = connections[DEFAULT_DB_ALIAS]
    connection.prepare_database()
    executor = MigrationExecutor(connection)
    targets = executor.loader.graph.leaf_nodes()
    missing_migrations = [mig[0] for mig in executor.migration_plan(targets)]
    return missing_migrations


# this bundles set up and migration management
def setup_django(
    isettings: InstanceSettings,
    deploy_migrations: bool = False,
    create_migrations: bool = False,
    configure_only: bool = False,
    init: bool = False,
):
    if IS_RUN_FROM_IPYTHON:
        os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = "true"

    import dj_database_url
    import django
    from django.conf import settings
    from django.core.management import call_command

    # configuration
    if not settings.configured:
        default_db = dj_database_url.config(
            default=isettings.db,
            conn_max_age=600,
            conn_health_checks=True,
        )
        DATABASES = {
            "default": default_db,
        }
        from .._init_instance import get_schema_module_name

        schema_names = ["core"] + list(isettings.schema)
        schema_module_names = [get_schema_module_name(n) for n in schema_names]

        settings.configure(
            INSTALLED_APPS=schema_module_names,
            DATABASES=DATABASES,
            DEFAULT_AUTO_FIELD="django.db.models.BigAutoField",
            TIME_ZONE="UTC",
            USE_TZ=True,
        )
        django.setup(set_prefix=False)

    if configure_only:
        return None

    # migrations management
    if create_migrations:
        call_command("makemigrations")
        return None

    # check that migrations have been deployed
    missing_migrations = get_migrations_to_sync()
    if len(missing_migrations) > 0:
        if deploy_migrations:
            verbosity = 0 if init else 2
            call_command("migrate", verbosity=verbosity)
            if not init:
                # only update if called from lamin migrate deploy
                # if called from load_schema(..., init=True)
                # no need to update the remote sqlite
                isettings._update_cloud_sqlite_file()
        else:
            logger.warning(
                MISSING_MIGRATIONS_WARNING.format(missing_migrations=missing_migrations)
            )
    else:
        if deploy_migrations:
            logger.success("database already up-to-date with migrations!")

    global IS_SETUP
    IS_SETUP = True



"""Setup & configure LaminDB.

Every function in the API matches a command in the `lamin` CLI.

Guide: :doc:`docs:setup`.

Setup:

.. autosummary::
   :toctree:

   login
   init
   load
   close
   delete

More instance operations:

.. autosummary::
   :toctree:

   info
   set
   migrate
   register

Settings:

.. autosummary::
   :toctree:

   settings

Developer API.

.. autosummary::
   :toctree:

   dev
"""


__version__ = "0.53.2"  # denote a release candidate for 0.1.0 with 0.1rc1

import sys
from os import name as _os_name

from . import dev
from ._check_instance_setup import check_instance_setup as _check_instance_setup  # noqa
from ._close import close  # noqa
from ._delete import delete  # noqa
from ._info import info  # noqa
from ._init_instance import init  # noqa
from ._load_instance import load  # noqa
from ._migrate import migrate
from ._register_instance import register  # noqa
from ._set import set, set_storage  # noqa
from ._settings import settings  # noqa
from ._setup_user import login, signup  # noqa

_TESTING = False  # used in lamindb tests

# hide the supabase error in a thread on windows
if _os_name == "nt":
    if sys.version_info.minor > 7:
        import threading

        _original_excepthook = threading.excepthook

        def _except_hook(args):
            is_overflow = args.exc_type is OverflowError
            for_timeout = str(args.exc_value) == "timeout value is too large"
            if not (is_overflow and for_timeout):
                _original_excepthook(args)

        threading.excepthook = _except_hook



from typing import Any, Dict, Iterable, List, Tuple, Union

import pandas as pd
from django.core.exceptions import FieldDoesNotExist
from lamin_utils import colors, logger
from lnschema_core.models import Feature, Registry, ULabel
from lnschema_core.types import ListLike, StrField

from .dev._settings import settings


# The base function for `from_values`
def get_or_create_records(
    iterable: ListLike,
    field: StrField,
    *,
    from_bionty: bool = False,
    **kwargs,
) -> List[Registry]:
    """Get or create records from iterables."""
    upon_create_search_names = settings.upon_create_search_names
    settings.upon_create_search_names = False
    feature: Feature = None
    try:
        Registry = field.field.model
        iterable_idx = index_iterable(iterable)

        # returns existing records & non-existing values
        records, nonexist_values, msg = get_existing_records(
            iterable_idx=iterable_idx, field=field, kwargs=kwargs
        )

        # new records to be created based on new values
        if len(nonexist_values) > 0:
            if from_bionty:
                records_bionty, unmapped_values = create_records_from_bionty(
                    iterable_idx=nonexist_values, field=field, msg=msg, **kwargs
                )
                if len(records_bionty) > 0:
                    msg = ""
                for record in records_bionty:
                    record._from_bionty = True
                records += records_bionty
            else:
                unmapped_values = nonexist_values
            # unmapped new_ids will NOT create records
            if len(unmapped_values) > 0:
                if len(msg) > 0:
                    logger.success(msg)
                s = "" if len(unmapped_values) == 1 else "s"
                print_values = colors.yellow(_print_values(unmapped_values))
                name = Registry.__name__
                n_nonval = colors.yellow(f"{len(unmapped_values)} non-validated")
                logger.warning(
                    f"{colors.red('did not create')} {name} record{s} for "
                    f"{n_nonval} {colors.italic(f'{field.field.name}{s}')}: {print_values}"  # noqa
                )
        if Registry.__module__.startswith("lnschema_bionty.") or Registry == ULabel:
            if isinstance(iterable, pd.Series):
                feature = iterable.name
            feature_name = None
            if isinstance(feature, str):
                feature_name = feature
            if feature_name is not None:
                if feature_name is not None:
                    for record in records:
                        record._feature = feature_name
                logger.debug(f"added default feature '{feature_name}'")
        return records
    finally:
        settings.upon_create_search_names = upon_create_search_names


def get_existing_records(
    iterable_idx: pd.Index,
    field: StrField,
    kwargs: Dict = {},
):
    model = field.field.model
    condition: Dict = {} if len(kwargs) == 0 else kwargs.copy()

    if _has_species_field(model):
        from lnschema_bionty._bionty import create_or_get_species_record

        species_record = create_or_get_species_record(
            species=kwargs.get("species"), orm=model
        )
        if species_record is not None:
            kwargs.update({"species": species_record})
            condition.update({"species": species_record})

    # standardize based on the DB reference
    # log synonyms mapped terms
    result = model.inspect(
        iterable_idx, field=field, species=kwargs.get("species"), mute=True
    )
    syn_mapper = result.synonyms_mapper

    syn_msg = ""
    if len(syn_mapper) > 0:
        s = "" if len(syn_mapper) == 1 else "s"
        names = list(syn_mapper.keys())
        print_values = colors.green(_print_values(names))
        syn_msg = (
            "loaded"
            f" {colors.green(f'{len(syn_mapper)} {model.__name__} record{s}')}"
            f" matching {colors.italic('synonyms')}: {print_values}"
        )
        iterable_idx = iterable_idx.to_frame().rename(index=syn_mapper).index

    # get all existing records in the db
    # if necessary, create records for the values in kwargs
    # k:v -> k:v_record
    # kwargs is used to deal with species
    condition.update({f"{field.field.name}__in": iterable_idx.values})

    query_set = model.filter(**condition)
    records = query_set.list()

    # now we have to sort the list of queried records
    # preserved = Case(
    #     *[
    #         When(**{field.field.name: value}, then=pos)
    #         for pos, value in enumerate(iterable_idx)
    #     ]
    # )
    # order by causes a factor 10 in runtime
    # records = query_set.order_by(preserved).list()

    # log validated terms
    validated = result.validated
    msg = ""
    if len(validated) > 0:
        s = "" if len(validated) == 1 else "s"
        print_values = colors.green(_print_values(validated))
        msg = (
            "loaded"
            f" {colors.green(f'{len(validated)} {model.__name__} record{s}')}"
            f" matching {colors.italic(f'{field.field.name}')}: {print_values}"
        )

    # no logging if all values are validated
    # logs if there are synonyms
    if len(syn_msg) > 0:
        if len(msg) > 0:
            logger.success(msg)
        logger.success(syn_msg)
        msg = ""

    existing_values = iterable_idx.intersection(
        query_set.values_list(field.field.name, flat=True)
    )
    nonexist_values = iterable_idx.difference(existing_values)

    return records, nonexist_values, msg


def create_records_from_bionty(
    iterable_idx: pd.Index,
    field: StrField,
    msg: str = "",
    **kwargs,
):
    model = field.field.model
    records: List = []
    # populate additional fields from bionty
    from lnschema_bionty._bionty import get_bionty_source_record

    # create the corresponding bionty object from model
    bionty_object = model.bionty(
        species=kwargs.get("species"), bionty_source=kwargs.get("bionty_source")
    )
    # add bionty_source record to the kwargs
    kwargs.update({"bionty_source": get_bionty_source_record(bionty_object)})

    # filter the columns in bionty df based on fields
    bionty_df = _filter_bionty_df_columns(model=model, bionty_object=bionty_object)

    # standardize in the bionty reference
    result = bionty_object.inspect(iterable_idx, field=field.field.name, mute=True)
    syn_mapper = result.synonyms_mapper

    msg_syn: str = ""
    if len(syn_mapper) > 0:
        s = "" if len(syn_mapper) == 1 else "s"
        names = list(syn_mapper.keys())
        print_values = colors.purple(_print_values(names))
        msg_syn = (
            "created"
            f" {colors.purple(f'{len(syn_mapper)} {model.__name__} record{s} from Bionty')}"  # noqa
            f" matching {colors.italic('synonyms')}: {print_values}"
        )

        iterable_idx = iterable_idx.to_frame().rename(index=syn_mapper).index

    # create records for values that are found in the bionty reference
    # matching either field or synonyms
    mapped_values = iterable_idx.intersection(bionty_df[field.field.name])

    multi_msg = ""
    if len(mapped_values) > 0:
        bionty_kwargs, multi_msg = _bulk_create_dicts_from_df(
            keys=mapped_values, column_name=field.field.name, df=bionty_df
        )
        for bk in bionty_kwargs:
            records.append(model(**bk, **kwargs))

        # number of records that matches field (not synonyms)
        validated = result.validated
        if len(validated) > 0:
            s = "" if len(validated) == 1 else "s"
            print_values = colors.purple(_print_values(validated))
            # this is the success msg for existing records in the DB
            if len(msg) > 0:
                logger.success(msg)
            logger.success(
                (
                    "created"
                    f" {colors.purple(f'{len(validated)} {model.__name__} record{s} from Bionty')}"  # noqa
                    f" matching {colors.italic(f'{field.field.name}')}: {print_values}"  # noqa
                )
            )

    # make sure that synonyms logging appears after the field logging
    if len(msg_syn) > 0:
        logger.success(msg_syn)
    # warning about multi matches
    if len(multi_msg) > 0:
        logger.warning(multi_msg)

    # return the values that are not found in the bionty reference
    unmapped_values = iterable_idx.difference(mapped_values)
    return records, unmapped_values


def index_iterable(iterable: Iterable) -> pd.Index:
    idx = pd.Index(iterable).unique()
    # No entries are made for NAs, '', None
    # returns an ordered unique not null list
    return idx[(idx != "") & (~idx.isnull())]


def _print_values(names: List, n: int = 20) -> str:
    print_values = ", ".join([f"'{name}'" for name in names[:n]])
    if len(names) > n:
        print_values += ", ..."
    return print_values


def _filter_bionty_df_columns(model: Registry, bionty_object: Any) -> pd.DataFrame:
    bionty_df = pd.DataFrame()
    if bionty_object is not None:
        model_field_names = {i.name for i in model._meta.fields}
        # parents needs to be added here as relationships aren't in fields
        model_field_names.add("parents")
        bionty_df = bionty_object.df().reset_index()
        if model.__name__ == "Gene":
            # groupby ensembl_gene_id and concat ncbi_gene_ids
            groupby_id_col = (
                "ensembl_gene_id" if "ensembl_gene_id" in bionty_df else "stable_id"
            )
            bionty_df.drop(
                columns=["hgnc_id", "mgi_id", "index"], errors="ignore", inplace=True
            )
            bionty_df.drop_duplicates([groupby_id_col, "ncbi_gene_id"], inplace=True)
            bionty_df["ncbi_gene_id"] = bionty_df["ncbi_gene_id"].fillna("")
            bionty_df = (
                bionty_df.groupby(groupby_id_col)
                .agg(
                    {
                        "symbol": "first",
                        "ncbi_gene_id": "|".join,
                        "biotype": "first",
                        "description": "first",
                        "synonyms": "first",
                    }
                )
                .reset_index()
            )
            bionty_df.rename(columns={"ncbi_gene_id": "ncbi_gene_ids"}, inplace=True)
        # rename definition to description for the lnschema_bionty
        bionty_df.rename(columns={"definition": "description"}, inplace=True)
        bionty_df = bionty_df.loc[:, bionty_df.columns.isin(model_field_names)]
    return bionty_df


def _bulk_create_dicts_from_df(
    keys: Union[set, List], column_name: str, df: pd.DataFrame
) -> Tuple[Dict, str]:
    """Get fields from a DataFrame for many rows."""
    multi_msg = ""
    if df.index.name != column_name:
        df = df.set_index(column_name).loc[list(keys)]
    if not df.index.is_unique:
        # return all records for multi-matches with a warning
        dup = df.index[df.index.duplicated()].unique().tolist()
        if len(dup) > 0:
            s = "" if len(dup) == 1 else "s"
            print_values = _print_values(dup)
            multi_msg = (
                f"ambiguous validation in Bionty for {len(dup)} record{s}:"
                f" {print_values}"
            )

    return df.reset_index().to_dict(orient="records"), multi_msg


def _has_species_field(orm: Registry) -> bool:
    try:
        orm._meta.get_field("species")
        return True
    except FieldDoesNotExist:
        return False



# Generated by Django 4.2.2 on 2023-08-22 16:43

import django.db.models.deletion
from django.db import migrations, models

import lnschema_core.models


class Migration(migrations.Migration):
    dependencies = [
        ("lnschema_core", "0015_file_initial_version_file_version"),
    ]

    operations = [
        migrations.AddField(
            model_name="dataset",
            name="input_of",
            field=models.ManyToManyField(related_name="input_datasets", to="lnschema_core.run"),
        ),
        migrations.AddField(
            model_name="dataset",
            name="run",
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.PROTECT, related_name="output_datasets", to="lnschema_core.run"),
        ),
        migrations.AddField(
            model_name="dataset",
            name="transform",
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.PROTECT, related_name="datasets", to="lnschema_core.transform"),
        ),
        migrations.CreateModel(
            name="DatasetLabel",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("dataset", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="lnschema_core.dataset")),
                ("feature", models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to="lnschema_core.feature")),
                ("label", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="lnschema_core.label")),
            ],
            options={
                "unique_together": {("dataset", "label")},
            },
            bases=(models.Model, lnschema_core.models.LinkORM),
        ),
        migrations.RemoveField(
            model_name="dataset",
            name="labels",
        ),
        migrations.AddField(
            model_name="dataset",
            name="labels",
            field=models.ManyToManyField(related_name="datasets", through="lnschema_core.DatasetLabel", to="lnschema_core.label"),
        ),
        migrations.AddField(
            model_name="feature",
            name="modality",
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.PROTECT, related_name="features", to="lnschema_core.modality"),
        ),
        migrations.AlterField(
            model_name="featureset",
            name="modality",
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.PROTECT, related_name="feature_sets", to="lnschema_core.modality"),
        ),
        migrations.AlterUniqueTogether(
            name="transform",
            unique_together=set(),
        ),
        migrations.RemoveField(
            model_name="transform",
            name="stem_id",
        ),
        migrations.AddField(
            model_name="transform",
            name="initial_version",
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.PROTECT, to="lnschema_core.transform"),
        ),
        migrations.AlterField(
            model_name="transform",
            name="version",
            field=models.CharField(db_index=True, default=None, max_length=10, null=True),
        ),
        migrations.AlterField(
            model_name="storage",
            name="root",
            field=models.CharField(db_index=True, default=None, max_length=255, unique=True),
        ),
        migrations.RunSQL("UPDATE lnschema_core_file SET suffix = '' WHERE suffix IS NULL"),
        migrations.AlterField(
            model_name="file",
            name="suffix",
            field=models.CharField(db_index=True, default=None, max_length=30),
        ),
    ]



from typing import Optional

from lnschema_core.models import Run, Transform


def __init__(run: Run, *args, **kwargs):
    if len(args) == len(run._meta.concrete_fields):
        super(Run, run).__init__(*args, **kwargs)
        return None
    # now we proceed with the user-facing constructor
    if len(args) > 1:
        raise ValueError("Only one non-keyword arg allowed: transform")
    transform: Transform = None
    if "transform" in kwargs or len(args) == 1:
        transform = kwargs.pop("transform") if len(args) == 0 else args[0]
    reference: Optional[str] = (
        kwargs.pop("reference") if "reference" in kwargs else None
    )
    reference_type: Optional[str] = (
        kwargs.pop("reference_type") if "reference_type" in kwargs else None
    )
    if transform is None:
        raise TypeError("Pass transform parameter")
    if transform._state.adding:
        raise ValueError("Please save transform record before creating a run")
    super(Run, run).__init__(
        transform=transform,
        reference=reference,
        reference_type=reference_type,
    )


Run.__init__ = __init__



from lamin_utils import logger

from ._settings import settings
from .dev._settings_store import current_instance_settings_file
from .dev._setup_bionty_sources import delete_bionty_sources_yaml
from .dev.cloud_sqlite_locker import clear_locker


def close(mute: bool = False) -> None:
    """Close existing instance.

    Returns `None` if succeeds, otherwise an exception is raised.
    """
    if current_instance_settings_file().exists():
        instance = settings.instance.identifier
        try:
            settings.instance._update_cloud_sqlite_file()
        except Exception as e:
            if isinstance(e, FileNotFoundError):
                logger.warning("did not find local cache file")
            elif isinstance(e, PermissionError):
                logger.warning("did not upload cache file - not enough permissions")
            else:
                raise e
        current_instance_settings_file().unlink()
        delete_bionty_sources_yaml()
        clear_locker()
        if not mute:
            logger.success(f"closed instance: {instance}")
    else:
        if not mute:
            logger.info("no instance loaded")



"""Open-source data platform for biology.

LaminDB assumes that data is stored in scalable storage backends, typically
array formats like parquet, zarr, HDF5, TileDB or DuckDB or simple files in
object storage.

LaminDB helps you manage these data using registries for metadata.

The two most important are:

.. autosummary::
   :toctree: .

   File
   Dataset

Four registries track provenance of data batches:

.. autosummary::
   :toctree: .

   Transform
   Run
   User
   Storage

Four registries validate & contextualize data batches:

.. autosummary::
   :toctree: .

   ULabel
   Feature
   FeatureSet
   Modality

Functional tools:

.. autosummary::
   :toctree: .

   track
   view
   save

Static classes & modules:

.. autosummary::
   :toctree: .

   settings
   setup
   schema
   dev

"""

__version__ = "0.54.2"  # denote a release candidate for 0.1.0 with 0.1rc1

import os as _os

import lamindb_setup as _lamindb_setup

# prints warning of python versions
from lamin_utils import py_version_warning as _py_version_warning
from lamindb_setup import _check_instance_setup
from lamindb_setup._check_instance_setup import _INSTANCE_NOT_SETUP_WARNING

_py_version_warning("3.8", "3.11")

_TESTING = _lamindb_setup._TESTING
_INSTANCE_SETUP = _check_instance_setup(from_lamindb=True)
# allow the user to call setup
from . import setup  # noqa


class InstanceNotSetupError(Exception):
    pass


def __getattr__(name):
    raise InstanceNotSetupError(
        f"{_INSTANCE_NOT_SETUP_WARNING}If you used the CLI to init or load an instance,"
        " please RESTART the python session (in a notebook, restart kernel)"
    )


# only import all other functionality if setup was successful
if _INSTANCE_SETUP:
    del InstanceNotSetupError
    del __getattr__  # delete so that imports work out
    from lnschema_core import (  # noqa
        Dataset,
        Feature,
        FeatureSet,
        File,
        Modality,
        Run,
        Storage,
        Transform,
        ULabel,
        User,
    )

    from . import _dataset  # noqa
    from . import _feature  # noqa
    from . import _feature_set  # noqa
    from . import _file  # noqa
    from . import _parents  # noqa
    from . import _registry  # noqa
    from . import _run  # noqa
    from . import _storage  # noqa
    from . import _transform  # noqa
    from . import _ulabel  # noqa
    from . import _validate  # noqa
    from . import dev  # noqa
    from . import schema  # noqa
    from ._delete import delete  # noqa
    from ._registry import select_backward as select  # noqa
    from ._save import save  # noqa
    from ._view import view  # noqa
    from .dev import _priors  # noqa
    from .dev._run_context import run_context  # noqa
    from .dev._settings import settings

    track = run_context._track  # noqa
    add = save  # backward compat
    settings.__doc__ = """Global :class:`~lamindb.dev.Settings`."""
